

#include <stdlib.h>
//#include <iostream.h>
extern "C" {
  int printf(char*, ...) ;
}
extern "C" {
  #include "mpi.h"
}
#include "Seq.h"
#include "Procs.h"
#include "Threads.h"
#include "Tags.h"
#include "ProcTree.h"
#include "DataExchange.h"



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `GetServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


GetServer :: GetServer(int _threadId, char* _defData) :
                 Server(_threadId),
                 defData(_defData) {
  ths->installServer(threadId, this) ;

  MPI_Irecv(getBuf, GET_BUF_SIZE, MPI_INT,
            MPI_ANY_SOURCE, GET, MPI_COMM_WORLD, ths->event(threadId)) ;
      // enable receipt of first `get' request

  state = ALPHA_0 ;
}

GetServer :: ~GetServer() {
//  MPI_Cancel(ths->event(threadId)) ;
//      // disable receipt of further `get' requests
  {
    // bug work-around---MPI_Cancel broken.  Send message to self instead
    int locId ;
    MPI_Status status ;

    MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
    MPI_Send(getBuf, GET_BUF_SIZE, MPI_INT, locId, GET, MPI_COMM_WORLD) ;

    MPI_Wait(ths->event(threadId), &status) ;
  }
}


void GetServer :: transition(MPI_Status* status) {
  switch(state) {
    case ALPHA_0 : {
      char *dstData, *data ;
      int size, offset ;
      unpackGetBuf(&dstData, &size, &data, &offset) ;

      dstId = status->MPI_SOURCE ;
      packTegBuf(dstData, size) ;
      packDatBuf(size, data, offset) ;

      MPI_Isend(tegBuf, TEG_BUF_SIZE, MPI_INT,
                dstId, TEG, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable send of `teg' acknowledgement

      state = ALPHA_1 ;
      break ;
    }
    case ALPHA_1 : {
      MPI_Isend(datBuf, datBufSize, MPI_CHAR,
                dstId, DAT, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable send of data

      state = ALPHA_2 ;
      break ;
    }
    case ALPHA_2 : {
      MPI_Irecv(getBuf, GET_BUF_SIZE, MPI_INT,
                MPI_ANY_SOURCE, GET, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable receipt of next `get' request

      state = ALPHA_0 ;
      break ;
    }
  }
}

void GetServer :: getLoc(char* data, int size,
                         char* srcData, int srcOffset) {
  // Local `get' operation.

  if(!srcData) srcData = defData ;
  memcpy(data, srcData + srcOffset, size) ;
}

// Private buffer packing and unpacking routines

void GetServer :: unpackGetBuf(char** dstData, int* size,
                               char** data, int* offset) {
  *dstData   = (char*) (getBuf [0]) ;
  *size      = getBuf [1] ;
  *data      = (char*) (getBuf [2]) ;
  *offset    = getBuf [3] ;
}

void GetServer :: packTegBuf(char* dstData, int size) {
  tegBuf [0] = (int) dstData ;
  tegBuf [1] = size ;
}

void GetServer :: packDatBuf(int size, char* data, int offset) {
  if(!data) data = defData ;
  datBuf     = data + offset ;
  datBufSize = size ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `GarServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


GarServer :: GarServer(int _threadId, char* _defData) :
                 Server(_threadId),
                 defData(_defData) {
  ths->installServer(threadId, this) ;

  MPI_Irecv(garBuf, GAR_BUF_SIZE, MPI_INT,
            MPI_ANY_SOURCE, GAR, MPI_COMM_WORLD, ths->event(threadId)) ;
      // enable receipt of first `gar' request

  state = ALPHA_0 ;
}

GarServer :: ~GarServer() {
//  MPI_Cancel(ths->event(threadId)) ;
//      // disable receipt of further `gar' requests
  {
    // bug work-around---MPI_Cancel broken.  Send message to self instead
    int locId ;
    MPI_Status status ;

    MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
    MPI_Send(garBuf, GAR_BUF_SIZE, MPI_INT, locId, GAR, MPI_COMM_WORLD) ;

    MPI_Wait(ths->event(threadId), &status) ;
  }
}

void GarServer :: transition(MPI_Status* status) {
  switch(state) {
    case ALPHA_0 : {
      dstId = status->MPI_SOURCE ;

      int srcOffset ;
      unpackGarBuf(garBuf, &blockSize, &rank, &dstData, &srcData, &srcOffset) ;
      if(!srcData) srcData = defData ;
      srcData += srcOffset ;

      gasBufSize = 3 * rank ;
      gasBuf     = new int [gasBufSize] ;

      exts    = gasBuf ;
      dstStrs = exts + rank ;
      srcStrs = dstStrs + rank ;

      MPI_Irecv(gasBuf, gasBufSize, MPI_INT,
		dstId, GAS, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable receipt of array shape

      state = ALPHA_1 ;
      break ;
    }
    case ALPHA_1 : {
      MPI_Isend(garBuf, GAR_BUF_SIZE, MPI_INT,
                dstId, RAG, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable send of `rag' acknowledgement

      state = ALPHA_2 ;
      break ;
    }
    case ALPHA_2 : {
      MPI_Isend(gasBuf, gasBufSize, MPI_INT,
                dstId, SAG, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable send of array shape

      state = ALPHA_3 ;
      break ;
    }
    case ALPHA_3 : {
      createArrayType(&arrayType, blockSize, rank, exts, srcStrs) ;

      MPI_Isend(srcData, 1, *arrayType, dstId, DAR, 
                MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable send of data

      state = ALPHA_4 ;
      break ;
    }
    case ALPHA_4 : {
      deleteArrayType(&arrayType, rank) ;
      delete gasBuf ;

      MPI_Irecv(garBuf, GAR_BUF_SIZE, MPI_INT,
                MPI_ANY_SOURCE, GAR, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable receipt of next `gar' request

      state = ALPHA_0 ;
      break ;
    }
  }
}


// Private buffer packing and unpacking routines

void GarServer :: unpackGarBuf(int* buffer, int* blockSize, int* rank,
			       char** dstData,
                               char** srcData, int* srcOffset) {
  *blockSize = buffer [0] ;
  *rank      = buffer [1] ;
  *dstData   = (char*) (buffer [2]) ;
  *srcData   = (char*) (buffer [3]) ;
  *srcOffset = buffer [4] ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `ParServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


ParServer :: ParServer(int _threadId, char* _defData) :
                 Server(_threadId),
                 defData(_defData) {
  ths->installServer(threadId, this) ;

  MPI_Irecv(parBuf, PAR_BUF_SIZE, MPI_INT,
            MPI_ANY_SOURCE, PAR, MPI_COMM_WORLD, ths->event(threadId)) ;
      // enable receipt of first `par' request

  state = ALPHA_0 ;
}

ParServer :: ~ParServer() {
//  MPI_Cancel(ths->event(threadId)) ;
//      // disable receipt of further `gar' requests
  {
    // bug work-around---MPI_Cancel broken.  Send message to self instead
    int locId ;
    MPI_Status status ;

    MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
    MPI_Send(parBuf, PAR_BUF_SIZE, MPI_INT, locId, PAR, MPI_COMM_WORLD) ;

    MPI_Wait(ths->event(threadId), &status) ;
  }
}

void ParServer :: transition(MPI_Status* status) {
  switch(state) {
    case ALPHA_0 : {
      srcId = status->MPI_SOURCE ;

      int dstOffset ;
      unpackParBuf(parBuf, &blockSize, &rank, &dstData, &dstOffset) ;
      if(!dstData) dstData = defData ;
      dstData += dstOffset ;

      pasBufSize = 2 * rank ;
      pasBuf     = new int [pasBufSize] ;

      exts    = pasBuf ;
      dstStrs = exts + rank ;

      MPI_Irecv(pasBuf, pasBufSize, MPI_INT,
		srcId, PAS, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable receipt of array shape

      state = ALPHA_1 ;
      break ;
    }
    case ALPHA_1 : {
      createArrayType(&arrayType, blockSize, rank, exts, dstStrs) ;

      MPI_Irecv(dstData, 1, *arrayType,
                srcId, DAR, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable receipt of data

      state = ALPHA_2 ;
      break ;
    }
    case ALPHA_2 : {
      MPI_Isend(parBuf, PAR_BUF_SIZE, MPI_INT,
                srcId, RAP, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable send of `rap' acknowledgement

      state = ALPHA_3 ;
      break ;
    }
    case ALPHA_3 : {
      deleteArrayType(&arrayType, rank) ;
      delete pasBuf ;

      MPI_Irecv(parBuf, PAR_BUF_SIZE, MPI_INT,
                MPI_ANY_SOURCE, PAR, MPI_COMM_WORLD, ths->event(threadId)) ;
          // enable receipt of next `par' request

      state = ALPHA_0 ;
      break ;
    }
  }
}


// Private buffer packing and unpacking routines

void ParServer :: unpackParBuf(int* buffer, int* blockSize, int* rank,
			       char** dstData, int* dstOffset) {
  *blockSize = buffer [0] ;
  *rank      = buffer [1] ;
  *dstData   = (char*) (buffer [2]) ;
  *dstOffset = buffer [3] ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `TegServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


TegServer :: TegServer(int _threadId) : Server(_threadId) {
  ths->installServer(threadId, this) ;

  n = 0 ;

  state = GAMMA_0 ;
}

void TegServer :: transition(MPI_Status* status) {
  switch(state) {
    case GAMMA_0 : {
      unpackTegBuf() ;
      MPI_Irecv(datBuf, datBufSize, MPI_CHAR,
		status->MPI_SOURCE, DAT, MPI_COMM_WORLD, ths->event(threadId)) ;
            // enable receipt of data

      state = GAMMA_1 ;
      break ;
    }
    case GAMMA_1 : {
      n-- ;
      if(n > 0) {
        MPI_Irecv(tegBuf, TEG_BUF_SIZE, MPI_INT,
                  MPI_ANY_SOURCE, TEG, MPI_COMM_WORLD, ths->event(threadId)) ;
            // enable receipt of next `teg' acknowledgement
      }

      state = GAMMA_0 ; 
      break ;
    }
  }
}

void TegServer :: notify() {
  // A `get' has been inititiated.

  // (This operation can be safely invoked, regardless of the state
  // of the `TegServer'---including between receipt of `teg' and `dat').

  if(n == 0)
    MPI_Irecv(tegBuf, TEG_BUF_SIZE, MPI_INT,
              MPI_ANY_SOURCE, TEG, MPI_COMM_WORLD, ths->event(threadId)) ;
        // enable receipt of `teg' acknowledgement

  n++ ;
}

int TegServer :: active() {
  return n ;
}


// Private buffer unpacking routines

void TegServer :: unpackTegBuf() {
  datBuf     = (char*) tegBuf [0] ;
  datBufSize = tegBuf [1] ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `RagServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


RagServer :: RagServer(int _threadId) : Server(_threadId) {
  ths->installServer(threadId, this) ;

  n = 0 ;

  state = GAMMA_0 ;
}

void RagServer :: transition(MPI_Status* status) {
  switch(state) {
    case GAMMA_0 : {
      srcId = status->MPI_SOURCE ;
      unpackRagBuf(ragBuf, &blockSize, &rank, &dstData) ;

      int sagBufSize = 3 * rank ;
      sagBuf = new int [sagBufSize] ;

      exts    = sagBuf ;
      dstStrs = exts + rank ;

      MPI_Irecv(sagBuf, sagBufSize, MPI_INT,
		srcId, SAG, MPI_COMM_WORLD, ths->event(threadId)) ;
            // enable receipt of array shape

      state = GAMMA_1 ;
      break ;
    }
    case GAMMA_1 : {
      createArrayType(&arrayType, blockSize, rank, exts, dstStrs) ;

      MPI_Irecv(dstData, 1, *arrayType,
		srcId, DAR, MPI_COMM_WORLD, ths->event(threadId)) ;
            // enable receipt of data

      state = GAMMA_2 ;
      break ;
    }
    case GAMMA_2 : {
      deleteArrayType(&arrayType, rank) ;
      delete sagBuf ;

      n-- ;
      if(n > 0) {
        MPI_Irecv(ragBuf, GAR_BUF_SIZE, MPI_INT,
                  MPI_ANY_SOURCE, RAG, MPI_COMM_WORLD, ths->event(threadId)) ;
            // enable receipt of next `rag' acknowledgement
      }

      state = GAMMA_0 ; 
      break ;
    }
  }
}

void RagServer :: notify() {
  // A `gar' has been inititiated.

  // (This operation can be safely invoked, regardless of the state
  // of the `RagServer'---including between receipt of `rag' and `sag',
  // or `sag' and `dar').

  if(n == 0)
    MPI_Irecv(ragBuf, GAR_BUF_SIZE, MPI_INT,
	      MPI_ANY_SOURCE, RAG, MPI_COMM_WORLD, ths->event(threadId)) ;
	// enable receipt of `rag' acknowledgement

  n++ ;
}

int RagServer :: active() {
  return n ;
}

// Private buffer unpacking routines

void RagServer :: unpackRagBuf(int* buffer, int* blockSize, int* rank,
                               char** dstData) {
  *blockSize = buffer [0] ;
  *rank      = buffer [1] ;
  *dstData   = (char*) (buffer [2]) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `RapServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


RapServer :: RapServer(int _threadId) : Server(_threadId) {
  ths->installServer(threadId, this) ;

  n = 0 ;
}

void RapServer :: transition(MPI_Status* status) {
  n-- ;
  if(n > 0) {
    MPI_Irecv(rapBuf, PAR_BUF_SIZE, MPI_INT,
	      MPI_ANY_SOURCE, RAP, MPI_COMM_WORLD, ths->event(threadId)) ;
	// enable receipt of next `rap' acknowledgement
  }
}

void RapServer :: notify() {
  // A `par' has been inititiated.

  if(n == 0)
    MPI_Irecv(rapBuf, PAR_BUF_SIZE, MPI_INT,
	      MPI_ANY_SOURCE, RAP, MPI_COMM_WORLD, ths->event(threadId)) ;
	// enable receipt of `rap' acknowledgement

  n++ ;
}

int RapServer :: active() {
  return n ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `DataExchange' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


DataExchange :: DataExchange(ProcTree* _spanTree, char* _defData) :
                    spanTree(_spanTree),
                    defData(_defData) {

  int baseId = ths->allocateServerGroup(6) ;

  getServer = new GetServer(baseId,     _defData) ;
  tegServer = new TegServer(baseId + 1) ;
  garServer = new GarServer(baseId + 2, _defData) ;
  ragServer = new RagServer(baseId + 3) ;
  parServer = new ParServer(baseId + 4, _defData) ;
  rapServer = new RapServer(baseId + 5) ;

  MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;

  spanTree->synch() ;
}


DataExchange :: ~DataExchange() {
  while(tegServer->active() || ragServer->active() || rapServer->active()) 
    ths->waitServerTransition() ;

  spanTree->synch() ;

  delete rapServer ;
  delete parServer ;
  delete ragServer ;
  delete garServer ;
  delete tegServer ;
  delete getServer ;

  ths->deallocateServerGroup() ;
}

void DataExchange :: get(char* data, int size,
                         int remId, char* remData, int remOffset) {

  // Issue request for data from process `remId'.

  // This returns asynchronously---`*data' will remain in an
  // undefined state until the barrier is crossed.

  if(remId == locId)
    getServer->getLoc(data, size, remData, remOffset) ;
  else {
    packGetBuf(data, size, remData, remOffset) ;
    MPI_Isend(getBuf, GET_BUF_SIZE, MPI_INT,
              remId, GET, MPI_COMM_WORLD, ths->event(0)) ;

    ths->waitEvent(0) ;

    tegServer->notify() ;  // tell `tegServer' a response is expected.
  }
}


void DataExchange :: getArray(int blockSize, int rank, int* exts,
                char* dstData, int* dstStrs,
                int srcId, char* srcData, int srcOffset, int* srcStrs) {

  // Issue request for array data from process `srcId'.

  // This returns asynchronously---`*dstData' will remain in an
  // undefined state until the barrier is crossed.

  if(srcId == locId) {
    if(!srcData) srcData = defData ;
    arrcpy(blockSize, rank, exts,
	   dstData, dstStrs, srcData + srcOffset, srcStrs) ;
  }
  else {
    int garBuf [GAR_BUF_SIZE] ;
    packGarBuf(garBuf, blockSize, rank, dstData, srcData, srcOffset) ;

    MPI_Isend(garBuf, GAR_BUF_SIZE, MPI_INT,
              srcId, GAR, MPI_COMM_WORLD, ths->event(0)) ;

    ths->waitEvent(0) ;

    const int gasBufSize = 3 * rank ;
    int* gasBuf = new int [gasBufSize] ;
    packGasBuf(gasBuf, rank, exts, dstStrs, srcStrs) ;

    MPI_Isend(gasBuf, gasBufSize, MPI_INT,
              srcId, GAS, MPI_COMM_WORLD, ths->event(0)) ;

    ths->waitEvent(0) ;

    delete gasBuf ;

    ragServer->notify() ;  // tell `ragServer' a response is expected.
  }
}


void DataExchange :: putArray(int blockSize, int rank, int* exts,
                char* srcData, int* srcStrs,
                int dstId, char* dstData, int dstOffset, int* dstStrs) {

  // Send array data to process `dstId'.

  if(dstId == locId) {
    if(!dstData) dstData = defData ;
    arrcpy(blockSize, rank, exts,
	   dstData + dstOffset, dstStrs, srcData, srcStrs) ;
  }
  else {
    int parBuf [PAR_BUF_SIZE] ;
    packParBuf(parBuf, blockSize, rank, dstData, dstOffset) ;

    MPI_Isend(parBuf, PAR_BUF_SIZE, MPI_INT,
              dstId, PAR, MPI_COMM_WORLD, ths->event(0)) ;

    ths->waitEvent(0) ;

    const int pasBufSize = 2 * rank ;
    int* pasBuf = new int [pasBufSize] ;
    packPasBuf(pasBuf, rank, exts, dstStrs) ;

    MPI_Isend(pasBuf, pasBufSize, MPI_INT,
              dstId, PAS, MPI_COMM_WORLD, ths->event(0)) ;

    ths->waitEvent(0) ;

    delete pasBuf ;

    MPI_Datatype* arrayType ;
    createArrayType(&arrayType, blockSize, rank, exts, srcStrs) ;

    MPI_Isend(srcData, 1, *arrayType,
	      dstId, DAR, MPI_COMM_WORLD, ths->event(0)) ;
	  // enable receipt of data

    ths->waitEvent(0) ;

    deleteArrayType(&arrayType, rank) ;

    rapServer->notify() ;  // tell `rapServer' a response is expected.
  }
}


// Private buffer packing routines

void DataExchange :: packGetBuf(char* data, int size,
                       char* remData, int remOffset) {
  getBuf [0] = (int) data ;
  getBuf [1] = size ;
  getBuf [2] = (int) remData ;
  getBuf [3] = remOffset ;
}

void DataExchange :: packGarBuf(int* buffer, int blockSize, int rank,
				char* dstData, char* srcData, int srcOffset) {
  buffer [0] = blockSize ;
  buffer [1] = rank ;
  buffer [2] = (int) dstData ;
  buffer [3] = (int) srcData ;
  buffer [4] = srcOffset ;
}

void DataExchange :: packGasBuf(int* buffer, int rank,
                                int* exts, int* dstStrs, int* srcStrs) {
  memcpy((char *) (buffer), (char *) exts, rank * sizeof(int)) ;
  memcpy((char *) (buffer + rank), (char *) dstStrs, rank * sizeof(int)) ;
  memcpy((char *) (buffer + 2 * rank), (char *) srcStrs, rank * sizeof(int)) ;
}


void DataExchange :: packParBuf(int* buffer, int blockSize, int rank,
				char* dstData, int dstOffset) {
  buffer [0] = blockSize ;
  buffer [1] = rank ;
  buffer [2] = (int) dstData ;
  buffer [3] = dstOffset ;
}

void DataExchange :: packPasBuf(int* buffer, int rank,
                                int* exts, int* dstStrs) {
  memcpy((char *) (buffer), (char *) exts, rank * sizeof(int)) ;
  memcpy((char *) (buffer + rank), (char *) dstStrs, rank * sizeof(int)) ;
}


// Private routine for copying a multi-dimensional array.

void DataExchange :: arrcpy(int blockSize, int rank, int* exts,
			    char* dstData, int* dstStrs,
			    char* srcData, int* srcStrs) {
  if(rank == 0)
    memcpy(dstData, srcData, blockSize) ;
  else
    for(int i = 0 ; i < exts [rank - 1] ; i++)
      arrcpy(blockSize, rank - 1, exts,
             dstData + i * dstStrs [rank - 1], dstStrs,
             srcData + i * srcStrs [rank - 1], srcStrs) ;
}


// ----------------------------------------------------------------------
// Things to do
//
//   Tidy implementation and interface of `get', to conform with practises
//       adopted in `getArray'.

